/*
 *  transas_fpga_v02_ide.c
 * 
 *  IDE controller driver for Transas custom chip.
 *
 *  Copyright (C) 2001 OKTET Ltd., St.-Petersburg, Russia
 *  Author: Eugeny S. Mints <Eugeny.Mints@oktet.ru>
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *  http://www.OARcorp.com/rtems/license.html.
 * 
 * $Id: transas_fpga_v02_ide.c,v 1.11 2002/09/25 13:33:58 jack Exp $
 *
 */
#include <assert.h>
#include <rtems/blkdev.h>
#include <libchip/ide_ctrl_cfg.h>
#include <libchip/ide_ctrl_io.h>

#include "plm.h"
#include "transas_fpga_v02_ide.h"


extern void mfd_enable_inl(int inlnum);
extern void mfd_disable_inl(int inlnum);


/*
 * transas_fpga_v02_ide_probe --
 *     Controller probe routine
 *
 * PARAMETERS:
 *     minor - controller minor number
 *
 * RETURNS:
 *     TRUE/FALSE
 */
rtems_boolean
transas_fpga_v02_ide_probe(int minor)
{
    return TRUE;
}

/*
 * transas_fpga_v02_ide_initialize --
 *    controller initialization routine
 *
 * PARAMETERS:
 *     minor - controller minor number
 *
 * RETURNS:
 *     NONE
 *
 */
static void
transas_fpga_v02_ide_initialize(int minor)
{
    ide_controller_bsp_table_t *ctrl = &IDE_Controller_Table[minor];

    /* enable interrutps if necessary */
    if (ctrl->int_driven == TRUE)
    {
        /* 
         * disable interrupts on ide chip - it  will be enable later by 
         * generic ATA driver 
         */
        TR_IDE_CTRL_REG_WRITE(ctrl->port1, IDE_REGISTER_DEVICE_CONTROL, 
                              IDE_REGISTER_DEVICE_CONTROL_nIEN);

        /* enable interrupts on fpga */
#if 0
        /* 
         * Temporary hack - because of firmware fpga bug. Will be 
         * uncommented in future
         */
        mfd_enable_inl(PLM_IDE_INL);                      
#endif        
    }        
}

/*
 * transas_fpga_v02_ide_reg_read --
 *    Read register of the controller
 *
 * PARAMETERS:
 *     minor - controller minor number
 *     reg   - register to read
 *     value - placeholder for result
 *
 * RETURNS:
 *     NONE
 *
 */
static void
transas_fpga_v02_ide_reg_read(int minor, int reg, unsigned16 *value)
{
    unsigned32 port = IDE_Controller_Table[minor].port1;
        
    assert(value);
    
    switch (reg)
    {
        case IDE_REGISTER_ALTERNATE_STATUS_OFFSET:
            TR_IDE_CTRL_REG8_READ(port, IDE_REGISTER_ALTERNATE_STATUS, *value);            
            break;
        case IDE_REGISTER_DATA_WORD:
            TR_IDE_CMD_REG16_READ(port, IDE_REGISTER_DATA, *value);
            break;
        case IDE_REGISTER_DATA_BYTE:
            TR_IDE_CMD_REG8_READ(port, IDE_REGISTER_DATA, *value);
            break;
        case IDE_REGISTER_ERROR:
        case IDE_REGISTER_SECTOR_COUNT:
        case IDE_REGISTER_SECTOR_NUMBER:
        case IDE_REGISTER_CYLINDER_LOW:
        case IDE_REGISTER_CYLINDER_HIGH:
        case IDE_REGISTER_DEVICE_HEAD:
        case IDE_REGISTER_STATUS:
            TR_IDE_CMD_REG8_READ(port, reg, *value);
            break;
        default:
            rtems_fatal_error_occurred(RTEMS_INTERNAL_ERROR);
            break;    
    }
}

/*
 * transas_fpga_v02_ide_reg_write --
 *    Write  register of the controller
 *
 * PARAMETERS:
 *     minor - controller minor number
 *     reg   - register to write
 *     value - data to write
 *
 * RETURNS:
 *     NONE
 *
 */
static void
transas_fpga_v02_ide_reg_write(int minor, int reg, unsigned16 value)
{
    unsigned32 port = IDE_Controller_Table[minor].port1;

    if (reg == IDE_REGISTER_DEVICE_CONTROL_OFFSET)
        TR_IDE_CTRL_REG_WRITE(port, IDE_REGISTER_DEVICE_CONTROL, value);
    else    
        TR_IDE_REG_WRITE(port, reg, value);
}

/* 
 * transas_fpga_v02_ide_read_block --
 *     Read data block via controller's data register
 *
 * PARAMETERS:
 *     minor      - minor number of controller
 *     block_size - number of bytes to read
 *     bufs       - set of buffers to store data
 *     cbuf       - number of current buffer from the set
 *     pos        - position inside current buffer 'cbuf'
 *
 * RETURNS:
 *     NONE
 */
static void
transas_fpga_v02_ide_read_block(int minor, unsigned16 block_size, 
                                blkdev_sg_buffer *bufs, 
                                rtems_unsigned32 *cbuf,
                                rtems_unsigned32 *pos)
{
    unsigned32  port = IDE_Controller_Table[minor].port1;
    unsigned8   byte;
    unsigned16  cnt = 0;
    unsigned16 *lbuf = (unsigned16 *)
                       ((unsigned8 *)(bufs[(*cbuf)].buffer) + (*pos));
    unsigned32  llength = bufs[(*cbuf)].length;
    
    assert(bufs);
    
    TR_IDE_CMD_REG8_READ(port, IDE_REGISTER_STATUS, byte);

    while ((byte & IDE_REGISTER_STATUS_DRQ) && 
           (cnt < block_size))
    {
        TR_IDE_CMD_REG16_READ(port, IDE_REGISTER_DATA, *lbuf);
        lbuf++;
        cnt += 2; (*pos) += 2;
        if ((*pos) == llength)
        {
            (*pos) = 0;
            (*cbuf)++;
            lbuf = bufs[(*cbuf)].buffer;
            llength = bufs[(*cbuf)].length;
        }
        TR_IDE_CMD_REG8_READ(port, IDE_REGISTER_STATUS, byte);
    } 
}

/* 
 * transas_fpga_v02_ide_write_block --
 *     Write data block via controller's data register
 *
 * PARAMETERS:
 *     minor      - minor number of controller
 *     block_size - number of bytes to write
 *     bufs       - set of buffers which store data
 *     cbuf       - number of current buffer from the set
 *     pos        - position inside current buffer 'cbuf'
 *
 * RETURNS:
 *     NONE
 */
static void
transas_fpga_v02_ide_write_block(int minor, unsigned16 block_size, 
                                 blkdev_sg_buffer *bufs,
                                 rtems_unsigned32 *cbuf,
                                 rtems_unsigned32 *pos)
{
    unsigned32  port = IDE_Controller_Table[minor].port1;
    unsigned8   byte;
    unsigned16  cnt = 0;
    unsigned16 *lbuf = (unsigned16 *)
                       ((unsigned8 *)(bufs[(*cbuf)].buffer) + (*pos));
    unsigned32  llength = bufs[(*cbuf)].length;
            
    assert(bufs);
    
    TR_IDE_CMD_REG8_READ(port, IDE_REGISTER_STATUS, byte);
    
    while ((byte & IDE_REGISTER_STATUS_DRQ) && 
           (cnt < block_size))
    {
        TR_IDE_REG_WRITE(port, IDE_REGISTER_DATA, *lbuf);
        lbuf++;
        cnt += 2; (*pos) += 2;
        if ((*pos) == llength)
        {
            (*pos) = 0;
            (*cbuf)++;
            lbuf = bufs[(*cbuf)].buffer;
            llength = bufs[(*cbuf)].length;
        }
        TR_IDE_CMD_REG8_READ(port, IDE_REGISTER_STATUS, byte);
    } 
}

/* 
 * transas_fpga_v02_ide_control -
 *     Control interface for controller
 *
 * PARAMETERS:
 *     minor - minor number of controller
 *     cmd   - command
 *     arg   - argument
 *
 * RETURNS:
 *     depends on 'cmd'
 */
static int 
transas_fpga_v02_ide_control(int minor, unsigned32 cmd, void *arg)
{
    return 0;
}

/* 
 * transas_fpga_v02_ide_config_io_speed -
 *     Set controller's speed of IO operations
 *
 * PARAMETERS:
 *     minor         - minor number of controller
 *     modes_avaible - speeds available
 *
 * RETURNS:
 *     RTEMS_SUCCESSFUL on success, or error code if
 *     error occured
 */
static rtems_status_code 
transas_fpga_v02_ide_config_io_speed(int minor, unsigned8 modes_avail)
{
    return RTEMS_SUCCESSFUL;
}

ide_ctrl_fns_t tr_ctrl_fns = {
    transas_fpga_v02_ide_probe,
    transas_fpga_v02_ide_initialize,
    transas_fpga_v02_ide_control,
    transas_fpga_v02_ide_reg_read,
    transas_fpga_v02_ide_reg_write,
    transas_fpga_v02_ide_read_block,
    transas_fpga_v02_ide_write_block,
    transas_fpga_v02_ide_config_io_speed      
};

